unit UBuilder;

{$mode delphi}

interface

uses
  Classes, SysUtils, UGrammer, StrUtils;

type

  { TBuilder }

  TBuilder = class
  private
    LocalFunctions: string;
    LocalFunctionIndex: Integer;
    function Indent(Count: Integer): string;
    function GetItemString(Item: TRuleItem; Required: Boolean; IndentLevel: Integer): string;
    function StringText(const Text: string; Escaped: Boolean): string;
    procedure BuildMain(FileName: string);
    procedure BuildParser(FileName: string);
    function BuildParserItems(Items: TRuleItems; Required: Boolean; IndentLevel: Integer): string;
    procedure BuildSource(FileName: string);
    procedure BuildTokenizer(FileName: string);
    function BuildTokenizerItems(SourceFile: TStrings; Items: TRuleItems): string;
  public
    Grammer: TGrammer;
    procedure BuildCompiler;
  end;


const
  BooleanText: array[Boolean] of string = ('False', 'True');

implementation

function TBuilder.Indent(Count: Integer): string;
begin
  Result := DupeString('  ', Count);
end;

function TBuilder.GetItemString(Item: TRuleItem; Required: Boolean; IndentLevel: Integer): string;
begin
  Result := '';
  case Item.RuleItemType of
    ritTerminal: Result := 'Expect(''' +
      StringText(Item.Terminal, Item.EscapedStrings) + ''', Required and ' + BooleanText[Required] + ')';
    ritNonTerminal: Result := 'Parse' + Item.NonTerminal.Name + '(Required and ' + BooleanText[Required] + ')';
    ritTerminalRange: Result := 'ExpectRange(''' +
      StringText(Item.TerminalFrom, Item.EscapedStrings) + ''', ''' +
      StringText(Item.TerminalTo, Item.EscapedStrings) + ''', Required and ' + BooleanText[Required] + ')';
    ritSubItems: begin
      LocalFunctions := LocalFunctions + 'function ParseSubitems' + IntToStr(LocalFunctionIndex) + '(Required: Boolean = False): Boolean;' + LineEnding;
      LocalFunctions := LocalFunctions + 'var' + LineEnding;
      LocalFunctions := LocalFunctions + '  OldPosition: TPosition;' + LineEnding;
      LocalFunctions := LocalFunctions + 'begin' + LineEnding;
      LocalFunctions := LocalFunctions + '  Result := True;' + LineEnding;
      LocalFunctions := LocalFunctions + BuildParserItems(Item.SubItems, Required, 1);
      LocalFunctions := LocalFunctions + 'end;' + LineEnding + LineEnding;

      Result := Result + 'ParseSubitems' + IntToStr(LocalFunctionIndex) + '(Required and ' + BooleanText[Required] + ')';
      Inc(LocalFunctionIndex);
    end;
  end;
end;

function TBuilder.StringText(const Text: string; Escaped: Boolean): string;
begin
  Result := Text;
  Result := StringReplace(Result, '''', '''''', [rfReplaceAll]);
  if Escaped then begin
    Result := StringReplace(Result, '\n', '''#10''', [rfReplaceAll]);
    Result := StringReplace(Result, '\r', '''#13''', [rfReplaceAll]);
    Result := StringReplace(Result, '\t', '''#9''', [rfReplaceAll]);
    Result := StringReplace(Result, '\\', '\', [rfReplaceAll]);
  end;
end;

procedure TBuilder.BuildParser(FileName: string);
var
  ParserFile: TStringList;
  Rule: TRule;
  FunctionBody: string;
begin
  ParserFile := TStringList.Create;
  with ParserFile do begin
    Add('unit Parser;');
    Add('');
    Add('{$MODE Delphi}');
    Add('');
    Add('interface');
    Add('');
    Add('uses');
    Add('  SysUtils, Tokenizer, Types;');
    Add('');
    Add('type');
    Add('  TPosition = record');
    Add('    Index: Integer;');
    Add('    Pos: TPoint;');
    Add('    procedure Calc(Text: string);');
    Add('  end;');
    Add('');
    Add('  TParser = class');
    Add('    Tokenizer: TTokenizer;');
    Add('    Content: string;');
    Add('    FileName: string;');
    Add('    Position: TPosition;');
    Add('    procedure Error(Text: string);');
    Add('    function Expect(Text: string; Required: Boolean = False): Boolean;');
    Add('    function ExpectRange(CharFrom, CharTo: Char; Required: Boolean = False): Boolean;');
    for Rule in Grammer.Rules do
    //if Rule.Level = rlParser then
      Add('    function Parse' + Rule.Name + '(Required: Boolean = False): Boolean;');
    Add('    constructor Create;');
    Add('    destructor Destroy; override;');
    Add('  end;');
    Add('');
    Add('');
    Add('implementation');
    Add('');
    Add('procedure TPosition.Calc(Text: string);');
    Add('var');
    Add('  I: Integer;');
    Add('begin');
    Add('  for I := 1 to Length(Text) do begin');
    Add('    Inc(Pos.X);');
    Add('    Inc(Index);');
    Add('    if Text[I] = LineEnding then begin');
    Add('      Pos.X := 1;');
    Add('      Inc(Pos.Y);');
    Add('    end;');
    Add('  end;');
    Add('end;');
    Add('');
    Add('constructor TParser.Create;');
    Add('begin');
    Add('  Position.Index := 1;');
    Add('  Position.Pos := Point(1, 1);');
    Add('  Tokenizer := TTokenizer.Create;');
    Add('end;');
    Add('');
    Add('destructor TParser.Destroy;');
    Add('begin');
    Add('  FreeAndNil(Tokenizer);');
    Add('end;');
    Add('');
    Add('procedure TParser.Error(Text: string);');
    Add('begin');
    Add('  WriteLn(FileName + ''('' + IntToStr(Position.Pos.X) + '','' + IntToStr(Position.Pos.Y) + '') Error: '' + Text);');
    Add('end;');
    Add('');
    Add('function TParser.Expect(Text: string; Required: Boolean = False): Boolean;');
    Add('var');
    Add('  ReadText: string;');
    Add('  StartPos: TPosition;');
    Add('begin');
    Add('  StartPos := Position;');
    Add('  ReadText := Copy(Content, Position.Index, Length(Text));');
    Add('  Position.Calc(Text);');
    Add('  Result := Text = ReadText;');
    Add('  if not Result then begin');
    Add('    Position := StartPos;');
    Add('    if Required then Error(''Expected "'' + Text + ''" but found "'' + ReadText + ''".'');');
    Add('  end else WriteLn(''Expect: '' + Text);');
    Add('end;');
    Add('');
    Add('function TParser.ExpectRange(CharFrom, CharTo: char; Required: Boolean = False): Boolean;');
    Add('var');
    Add('  ReadChar: Char;');
    Add('  StartPos: TPosition;');
    Add('begin');
    Add('  StartPos := Position;');
    Add('  ReadChar := Content[Position.Index];');
    Add('  Position.Calc(Content[Position.Index]);');
    Add('  Result := (ReadChar >= CharFrom) and (ReadChar <= CharTo);');
    Add('  if not Result then begin');
    Add('    Position := StartPos;');
    Add('    if Required then Error(''Expected "'' + CharFrom + ''" to "'' + CharTo + ''" but found "'' + ReadChar + ''".'');');
    Add('  end else WriteLn(''ExpectRange: '' + CharFrom + ''..'' + CharTo + '': '' + ReadChar);');
    Add('end;');
    Add('');
    for Rule in Grammer.Rules do
    //if Rule.Level = rlParser then
    begin
      LocalFunctionIndex := 1;
      LocalFunctions := '';
      FunctionBody := BuildParserItems(Rule.Items, True, 1);
      Text := Text + 'function TParser.Parse' + Rule.Name + '(Required: Boolean = False): Boolean;' + LineEnding;
      Text := Text + LocalFunctions;
      Text := Text + 'var' + LineBreak;
      Text := Text + '  OldPosition: TPosition;' + LineEnding;
      Text := Text + 'begin' + LineEnding;
      //Text := Text + '  WriteLn(''Rule check: ' + Rule.Name + ''');';
      Text := Text + '  Result := True;' + LineEnding;
      Text := Text + FunctionBody;
      Text := Text + '  if Result then WriteLn(''Rule: ' + Rule.Name + ''');';
      Text := Text + 'end;' + LineEnding + LineEnding;
    end;
    Add('');
    Add('end.');
    SaveToFile(FileName);
  end;
  FreeAndNil(ParserFile);
end;

function TBuilder.BuildParserItems(Items: TRuleItems; Required: Boolean; IndentLevel: Integer): string;
var
  I: Integer;
  Item: TRuleItem;
  ItemRequired: Boolean;
  Line: string;
begin
  Result := Indent(IndentLevel) + 'OldPosition := Position;' + LineEnding;
  I := 0;
  for Item in Items do begin
    ItemRequired := not Item.Optional and Required;
    Line := '';
    case Items.RuleType of
      rtOr: begin
        Line := Line + Indent(IndentLevel);
        if I > 0 then Line := Line + 'else ';
        Line := Line + 'if ' + GetItemString(Item, False, IndentLevel) +
          ' then Exit';
      end;
      rtAnd: begin
        if not Item.Optional then
          Line := Line + Indent(IndentLevel) + 'Result := Result and ';
        if Item.Repetitive then begin
          if not Item.Optional then
            Line := Line + Indent(IndentLevel) + GetItemString(Item, ItemRequired, IndentLevel) + ';' + LineEnding;
          Line := Line + Indent(IndentLevel) + 'repeat' + LineEnding;
          Inc(IndentLevel);
          Line := Line + Indent(IndentLevel) + 'if not ';
          ItemRequired := False;
        end else
          if Item.Optional then Line := Line + Indent(IndentLevel);
        Line := Line + GetItemString(Item, ItemRequired, IndentLevel);
        if Item.Repetitive then begin
          Dec(IndentLevel);
          Line := Line + ' then Break;' + LineEnding +
            Indent(IndentLevel) + 'until False';
        end;
        Line := Line + ';' + LineEnding +
          Indent(IndentLevel) + 'if not Result then begin' + LineEnding +
          Indent(IndentLevel + 1) + 'Position := OldPosition;' + LineEnding +
          Indent(IndentLevel + 1) + 'Exit;' + LineEnding +
          Indent(IndentLevel) + 'end;';

        // Check white spaces
        if Assigned(Items.Grammer.WhiteSpaceRule) and Items.WithWhiteSpeaces
        then begin
          Line := Line + Indent(IndentLevel) + 'repeat' + LineEnding;
          Inc(IndentLevel);
          Line := Line + Indent(IndentLevel) + 'if not Parse' +
            Items.Grammer.WhiteSpaceRule.Name + '(False)' +
            ' then Break;' + LineEnding;
          Dec(IndentLevel);
          Line := Line + Indent(IndentLevel) + 'until False;' + LineEnding;
        end;
      end;
    end;
    Result := Result + Line + LineEnding;
    Inc(I);
  end;
  case Items.RuleType of
    rtOr: begin
      Result := Result + Indent(IndentLevel) + 'else begin' + LineEnding;
      Result := Result + Indent(IndentLevel) + '  if Required then Error(''Unexpected token'');' + LineEnding;
      Result := Result + Indent(IndentLevel) + '  Result := False;' + LineEnding;
      Result := Result + Indent(IndentLevel) + 'end;' + LineEnding;
    end;
  end;
end;

procedure TBuilder.BuildMain(FileName: string);
var
  ParserFile: TStringList;
begin
  ParserFile := TStringList.Create;
  with ParserFile do begin
    Add('program Compiler;');
    Add('');
    Add('{$MODE Delphi}');
    Add('');
    Add('uses');
    Add('  Source, Parser, Tokenizer, SysUtils;');
    Add('');
    Add('procedure Compile(FileName: string);');
    Add('var');
    Add('  SourceFile: file of Char;');
    Add('  Parser: TParser;');
    Add('  I: Integer;');
    Add('begin');
    Add('  AssignFile(SourceFile, FileName);');
    Add('  Reset(SourceFile);');
    Add('  Parser := TParser.Create;');
    Add('  Parser.FileName := ExtractFileName(FileName);');
    Add('  SetLength(Parser.Content, FileSize(SourceFile));');
    Add('  I := 1;');
    Add('  while not Eof(SourceFile) do begin');
    Add('    Read(SourceFile, Parser.Content[I]);');
    Add('    Inc(I);');
    Add('  end;');
    Add('  CloseFile(SourceFile);');
    if Assigned(Grammer.TopRule) then
      Add('  Parser.Parse' + Grammer.TopRule.Name + '(True);');
    Add('  FreeAndNil(Parser);');
    Add('end;');
    Add('');
    Add('begin');
    Add('  if ParamCount > 0 then');
    Add('    Compile(ParamStr(1))');
    Add('    else WriteLn(''File name not specified as parameter.'');');
    Add('end.');
    SaveToFile(FileName);
  end;
  FreeAndNil(ParserFile);
end;

procedure TBuilder.BuildSource(FileName: string);
var
  Rule: TRule;
  Item: TRuleItem;
  SourceFile: TStringList;
  TypeSectionStarted: Boolean;
  I: Integer;
begin
  SourceFile := TStringList.Create;
  with SourceFile do begin
    Add('unit Source;');
    Add('');
    Add('{$MODE Delphi}');
    Add('');
    Add('interface');
    Add('');
    Add('uses');
    Add('  fgl;');
    Add('');
    TypeSectionStarted := False;
    for Rule in Grammer.Rules do
    //if Rule.CreateSourceNode then
    begin
      if not TypeSectionStarted then begin
        Add('type');
        TypeSectionStarted := True;
      end;
      Add('  T' + Rule.Name + ' = class;');
    end;
    Add('');
    for Rule in Grammer.Rules do
    //if Rule.CreateSourceNode then
    begin
      Add('  T' + Rule.Name + ' = class');
      I := 1;
      for Item in Rule.Items do begin
        if Item.RuleItemType = ritNonTerminal then
          if Item.Repetitive then
          Add('    ' + Item.NonTerminal.Name + IntToStr(I) + ': TFPGObjectList<T' + Item.NonTerminal.Name + '>;')
          else Add('    ' + Item.NonTerminal.Name + IntToStr(I) + ': T' + Item.NonTerminal.Name + ';');
        Inc(I);
      end;
      Add('  end;' + LineEnding);
    end;
    Add('');
    Add('implementation');
    Add('');

    Add('end.');

    SaveToFile(FileName);
  end;
  FreeAndNil(SourceFile);
end;

function TBuilder.BuildTokenizerItems(SourceFile: TStrings; Items: TRuleItems): string;
var
  Item: TRuleItem;
  Line: string;
begin
  Result := '';
  Line := '';
  with SourceFile do begin
    for Item in Items do begin
      case Item.RuleItemType of
        ritTerminal: Line := Line + '(Value = ''' + StringText(Item.Terminal, Item.EscapedStrings) + ''')';
        ritNonTerminal: Line := Line + 'Is' + Item.NonTerminalName + '(Value)';
        ritTerminalRange: Line := Line + '((Value > ''' + StringText(Item.TerminalFrom, Item.EscapedStrings) + ''') and (Value < ''' + StringText(Item.TerminalTo, Item.EscapedStrings) + '''))';
        ritSubItems: Line := Line + '(' + BuildTokenizerItems(SourceFile, Item.SubItems) + ')';
      end;
      if Item <> Items[Items.Count - 1] then
        Line := Line + ' ' + RuleTypeText[Items.RuleType] + ' ';
      if Length(Line) > 80 then begin
        Result := Result + Line + LineEnding;
        Line := '';
      end;
    end;
    Result := Result + Line;
  end;
end;

procedure TBuilder.BuildTokenizer(FileName: string);
var
  Rule: TRule;
  SourceFile: TStringList;
begin
  SourceFile := TStringList.Create;
  with SourceFile do begin
    Add('unit Tokenizer;');
    Add('');
    Add('{$MODE Delphi}');
    Add('');
    Add('interface');
    Add('');
    Add('uses');
    Add('  fgl;');
    Add('');
    Add('type');
    Add('  TTokenizer = class');
    Add('    function ReadNext: string;');
    Add('    function Expect(Text: string): Boolean;');
    for Rule in Grammer.Rules do
    if (Rule.Level = rlTokenizer) then begin
      Add('    function Is' + Rule.Name + '(Value: Char): Boolean;');
    end;
    Add('  end;');
    Add('');
    Add('implementation');
    Add('');
    Add('function TTokenizer.ReadNext: string;');
    Add('begin');
    Add('end;');
    Add('');
    Add('function TTokenizer.Expect(Text: string): Boolean;');
    Add('begin');
    Add('end;');


    for Rule in Grammer.Rules do
    if Rule.Level = rlTokenizer then begin
      Add('function TTokenizer.Is' + Rule.Name + '(Value: Char): Boolean;');
      Add('begin');
      Add('  Result := ' + BuildTokenizerItems(SourceFile, Rule.Items) + ';');
      Add('end;' + LineEnding);
    end;
    Add('end.');

    SaveToFile(FileName);
  end;
  FreeAndNil(SourceFile);
end;

procedure TBuilder.BuildCompiler;
var
  OutputDir: string;
begin
  OutputDir := 'Generated';
  ForceDirectories(OutputDir);

  BuildMain(OutputDir + DirectorySeparator + 'Compiler.pas');
  BuildSource(OutputDir + DirectorySeparator + 'Source.pas');
  BuildParser(OutputDir + DirectorySeparator + 'Parser.pas');
  BuildTokenizer(OutputDir + DirectorySeparator + 'Tokenizer.pas');
end;


end.

